from __future__ import unicode_literals
import datetime
import re
import os
from contextlib import contextmanager

from .browser import Browser, ParseError, Soup
from .browser import NoSuchElementException, \
    WebDriverException, \
    NoSuchWindowException


class WishlistElement(Soup):
    """Wishlist.get() returns an instance of this object"""
    @property
    def host(self):
        return os.environ.get("WISHLIST_HOST", "https://www.amazon.com")

    @property
    def uuid(self):
        m = re.search("/dp/([^/]+)", self.url)
        return m.group(1) if m else ""

    @property
    def url(self):
        href = ""
        el = self.soup.find("a", id=re.compile("^itemName_"))
        if el and ("href" in el.attrs):
            m = re.search("/dp/([^/]+)", el.attrs["href"])
            if m:
                href = "{}/dp/{}/".format(self.host, m.group(1))
                tag = os.environ.get("WISHLIST_REFERRER", "marcyescom-20")
                if tag:
                    href += "?tag={}".format(tag)

        return href

    @property
    def image(self):
        src = ""
        imgs = self.soup.find_all("img")
        for img in imgs:
            if "src" in img.attrs:
                if img.parent and img.parent.name == "a":
                    a = img.parent
                    if a.attrs["href"].startswith("/dp/"):
                        src = img.attrs["src"]
                        break

        if not src:
            for img in imgs:
                maybe_src = img.attrs.get("src", "")
                if "/images/I/" in maybe_src:
                    src = maybe_src
                    break

        return src

    @property
    def price(self):
        price = 0.0
        el = self.soup.find("span", id=re.compile("^itemPrice_"))
        if el and len(el.contents) > 0:
            try:
                price_str = el.contents[0].strip()
                if price_str:
                    price = float(price_str[1:].split()[0].replace(",", ""))
            except (ValueError, IndexError):
                price = 0.0

        return price

    @property
    def marketplace_price(self):
        price = 0.0
        el = self.soup.find("span", {"class": "itemUsedAndNewPrice"})
        if el and len(el.contents) > 0:
            price = float(el.contents[0].replace("$", "").replace(",", ""))
        return price

    @property
    def title(self):
        title = ""
        el = self.soup.find("a", id=re.compile("^itemName_"))
        if el and len(el.contents) > 0:
            title = el.contents[0].strip()

        else:
            el = self.soup.find("span", id=re.compile("^itemName_"))
            if el and len(el.contents) > 0:
                title = el.contents[0].strip()

        return title

    @property
    def comment(self):
        ret = ""
        el = self.soup.find("span", id=re.compile("^itemComment_"))
        if el and len(el.contents) > 0:
            ret = el.contents[0].strip()
        return ret

    @property
    def rating(self):
        stars = 0.0
        el = self.soup.find("a", {"class": "reviewStarsPopoverLink"})
        if el:
            el = el.find("span", {"class": "a-icon-alt"})
            if len(el.contents) > 0:
                stars = float(el.contents[0].strip().split()[0])
        return stars

    @property
    def author(self):
        author = ""
        el = self.soup.find("a", id=re.compile("^itemName_"))
        if el:
            author = el.parent.next_sibling
            if author:
                author = author.strip().replace("by ", "")
        return author

    @property
    def added(self):
        ret = None
        el = self.soup.find("div", id=re.compile("^itemAction_"))
        el = el.find("span", {"class": "a-size-small"})
        if el and len(el.contents) > 0:
            ret = el.contents[0].strip().replace("Added ", "")
            if ret:
                ret = datetime.datetime.strptime(ret, '%B %d, %Y')
        return ret

    @property
    def body(self):
        return self.soup.prettify()

    def __init__(self, element):
        self.soup = self.soupify(element)

    def jsonable(self):
        json_item = {}
        json_item["title"] = self.title
        json_item["image"] = self.image
        json_item["uuid"] = self.uuid
        json_item["url"] = self.url
        json_item["price"] = self.price
        json_item["marketplace_price"] = self.marketplace_price
        json_item["comment"] = self.comment
        json_item["author"] = self.author
        json_item["added"] = self.added.strftime('%B %d, %Y')
        json_item["rating"] = self.rating
        return json_item


class Wishlist(Browser):
    """Wrapper that is specifically designed for getting amazon wishlists"""

    element_class = WishlistElement

    @property
    def host(self):
        return os.environ.get("WISHLIST_HOST", "https://www.amazon.com")

    @classmethod
    @contextmanager
    def lifecycle(cls):
        with super(Wishlist, cls).lifecycle() as instance:
            instance.homepage() # we load homepage to force cookie loading
            yield instance

    @classmethod
    @contextmanager
    def open(cls):
        with super(Wishlist, cls).open() as instance:
            instance.homepage() # we load homepage to force cookie loading
            instance.current_page = 0
            yield instance

    def get_items_from_body(self, body):
        """this will return the wishlist elements on the current page"""
        soup = self.soupify(body)
        html_items = soup.findAll("div", {"id": re.compile("^item_")})
        for i, html_item in enumerate(html_items):
            item = self.element_class(html_item)
            yield item

    def get_total_pages_from_body(self, body):
        """return the total number of pages of the wishlist

        body -- string -- the complete html page
        """
        page = 0
        soup = self.soupify(body)

        try:
            #el = soup.find("ul", id=re.compile("^itemAction_"))
            el = soup.find("ul", {"class": "a-pagination"})
            #el = el.find("li", {"class": "a-last"})
            els = el.findAll("li", {"class": re.compile("^a-")})
            #pout.v(len(els))
            #pout.v(els[-2])
            el = els[-2]
            if len(el.contents) and len(el.contents[0].contents):
                page = int(el.contents[0].contents[0].strip())

        except AttributeError:
            raise ParseError("Could not find pagination, is this a wishlist page?")

        return page

    def get_wishlist_url(self, name, page):
        base_url = "{}/gp/registry/wishlist/{}".format(self.host, name)
        if page > 1:
            base_url += "?page={}".format(page)
        return base_url

    def get(self, name, page=0):
        """return the items of the given wishlist name"""

        crash_count = 0
        page = page if page > 1 else 1
        self.current_page = page

        soup = None
        page_count = None

        while True:
            try:
                # https://www.amazon.com/gp/registry/wishlist/NAME
                self.location(self.get_wishlist_url(name, page))
                soup = self.soupify(self.body)

                if page_count is None:
                    page_count = self.get_total_pages_from_body(soup)

                for i, item in enumerate(self.get_items_from_body(soup)):
                    yield item

            finally:
                page += 1
                if page_count is None or page > page_count:
                    break

                self.current_page = page
                soup = None

    def homepage(self, **kwargs):
        """loads the amazon homepage, this forces cookies to load"""
        self.location(self.host, **kwargs)




import logging

from .core import Wishlist, WishlistElement, ParseError


__version__ = "0.0.8"


# get rid of "No handler found" warnings (cribbed from requests)
logging.getLogger(__name__).addHandler(logging.NullHandler())




from __future__ import unicode_literals
import os
import tempfile
import sys
import codecs
import pickle
import logging
import time
from contextlib import contextmanager
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from selenium import webdriver
#from selenium.webdriver.common.keys import Keys
from selenium.common.exceptions import NoSuchElementException, \
    WebDriverException, \
    NoSuchWindowException
from pyvirtualdisplay import Display
from bs4 import BeautifulSoup
from bs4 import Tag

from selenium.webdriver.firefox.webdriver import WebDriver as BaseWebDriver
#from selenium.webdriver.chrome.options import Options
#from selenium.webdriver.firefox.webelement import FirefoxWebElement as BaseWebElement # selenium 3.0
#from selenium.webdriver.remote.webelement import WebElement as BaseWebElement # selenium <3.0

from .compat import *


logger = logging.getLogger(__name__)


class ParseError(RuntimeError):
    """This gets raised any time Browser.element() fails to parse,
    it wraps NoSuchElementException"""
    def __init__(self, body, e):
        self.body = body
        self.error = e
        super(ParseError, self).__init__(e.message)


class RecoverableCrash(IOError):
    def __init__(self, e):
        self.error = e
        super(RecoverableCrash, self).__init__(e.message)


class Soup(object):
    def soupify(self, body):
        if isinstance(body, Tag): return body
        soup = BeautifulSoup(body, "html.parser")
        return soup


# class WebElement(BaseWebElement):
#     @property
#     def body(self):
#         """Return the body of this particular element"""
#         # http://stackoverflow.com/a/22227980/5006
#         return self.get_attribute('innerHTML')
# 
#     @property
#     def soup(self):
#         """Return a beautiful soup object with the contents of this element"""
#         soup = getattr(self, "_soup", None)
#         if soup is None:
#             soup = BeautifulSoup(self.body, "html.parser")
#             self._soup = soup
#         return soup
# 
# 
# class WebDriver(BaseWebDriver):
#     _web_element_cls = WebElement # selenium 3.0
# 
#     def create_web_element(self, element_id):
#         # TODO -- remove when upgrading to selenium 3.0
#         return WebElement(self, element_id, w3c=self.w3c)


class Cookies(object):
    """This will write and read cookies for a Browser instance, calling Browser.location()
    will use this class to load cookies for the given domain if they are available"""
    @property
    def directory(self):
        directory = getattr(self, "_directory", None)
        if directory is None:
            directory = os.environ.get("BROWSER_CACHE_DIR", "")
            if directory:
                directory = os.path.abspath(os.path.expanduser(directory))
            else:
                directory = tempfile.gettempdir()

            self._directory = directory
        return directory

    @property
    def path(self):
        cookies_d = self.directory
        cookies_f = os.path.join(cookies_d, "{}.txt".format(self.domain))
        return cookies_f

    def __iter__(self):
        cookies_f = self.path
        if os.path.isfile(cookies_f):
            with open(cookies_f, "rb") as f:
                cookies = pickle.load(f)

            for cookie in cookies:
                yield cookie

    def save(self, cookies):
        """save the cookies in browser"""
        cookies_f = self.path
        with open(cookies_f, "w+b") as f:
            pickle.dump(cookies, f)

    def __init__(self, domain):
        """
        browser -- selenium web driver -- usually Firefox or Chrome
        """
        self.domain = domain


class Browser(Soup):
    """This is a wrapper around selenium and pyvirtualdisplay to make browsering
    from the command line easier

    link -- Selinium source -- https://github.com/SeleniumHQ/selenium/tree/master/py
    """
    @property
    def body(self):
        """return the body of the current page"""
        # http://stackoverflow.com/a/7866938/5006
        return self.browser.page_source

    @property
    def current_url(self):
        """return the current url"""
        # http://stackoverflow.com/questions/15985339/how-do-i-get-current-url-in-selenium-webdriver-2-python
        return self.browser.current_url

    @property
    def browser(self):
        """wrapper around the browser in case we want to switch from Firefox, you
        should use this over the .firefox property"""
        browser = getattr(self, "_browser", None)
        if browser is None:
            browser = self.chrome
            #browser._web_element_cls = WebElement
            self._browser = browser

        return browser

    @browser.deleter
    def browser(self):
        try:
            self._browser.close()
            del self._browser
        except (WebDriverException, AttributeError):
            pass

    @property
    def display(self):
        display = getattr(self, "_display", None)
        if display is None:
            # http://coreygoldberg.blogspot.com/2011/06/python-headless-selenium-webdriver.html
            display = Display(visible=0, size=(800, 600))
            display.start()
            self._display = display
        return display

    @property
    def chrome(self):
        chrome = webdriver.Chrome()
        return chrome

#     @property
#     def firefox(self):
#         profile = webdriver.FirefoxProfile()
#         #firefox = webdriver.Firefox(firefox_profile=profile)
#         firefox = WebDriver(firefox_profile=profile)
#         return firefox


    @classmethod
    @contextmanager
    def open(cls):
        """Where all the magic happens, you use this to start the virtual display
        and power up the browser

        with Browser.open() as browser:
            browser.location("http://example.com")
        """
        try:
            instance = cls()
            # start up the display
            instance.display
            yield instance

        except WebDriverException as e:
            logger.exception(e)
            del instance.browser
            raise RecoverableCrash(e)

        except Exception as e:
            logger.exception(e)
            exc_info = sys.exc_info()

            if instance:
                try:
                    directory = tempfile.gettempdir()
                    filename = os.path.join(directory, "wishlist.png")
                    instance.browser.get_screenshot_as_file(filename)
                except Exception as e:
                    pass

                try:
                    with codecs.open(os.path.join(directory, "wishlist.html"), encoding='utf-8', mode='w+') as f:
                        f.write(instance.body)
                except Exception as e:
                    pass

            reraise(*exc_info)

        finally:
            instance.close()






    # DEPRECATED 9-7-2016, use open() instead
    @classmethod
    @contextmanager
    def lifecycle(cls):
        """Where all the magic happens, you use this to start the virtual display
        and power up the browser

        with Browser.lifecycle() as browser:
            browser.location("http://example.com")
        """
        try:
            instance = cls()
            # start up the display
            instance.display
            yield instance

        except Exception as e:
            logger.exception(e)
            exc_info = sys.exc_info()

            if instance:
                try:
                    directory = tempfile.gettempdir()
                    filename = os.path.join(directory, "wishlist.png")
                    instance.browser.get_screenshot_as_file(filename)
                except Exception as e:
                    pass

                try:
                    with codecs.open(os.path.join(directory, "wishlist.html"), encoding='utf-8', mode='w+') as f:
                        f.write(instance.body)
                except Exception as e:
                    pass

            reraise(*exc_info)

        finally:
            instance.close()

    def location(self, url, ignore_cookies=False):
        """calls the selenium driver's .get() method, and will load cookies if they
        are available

        url -- string -- the full url (scheme, domain, path)
        ignore_cookies -- boolean -- if True then don't try and load cookies
        """
        logger.debug("Loading location {}".format(url))
        driver = self.browser
        driver.get(url)
        url_bits = urlparse.urlparse(url)
        domain = url_bits.hostname
        self.domain = domain

        if not ignore_cookies:
            if domain and (domain not in self.domains):
                logger.debug("Loading cookies for {}".format(domain))
                self.domains[domain] = domain
                cookies = Cookies(domain)
                count = 0
                for count, cookie in enumerate(cookies, 1):
                    driver.add_cookie(cookie)
                logger.debug("Loaded {} cookies".format(count))

    def element_exists(self, css_selector):
        ret = True
        try:
            self.browser.find_element_by_css_selector(css_selector)
        except NoSuchElementException as e:
            ret = False
        return ret

    def element(self, css_selector):
        """wrapper around Selenium's css selector that raises ParseError if fails"""
        try:
            return self.browser.find_element_by_css_selector(css_selector)

        except NoSuchElementException as e:
            logger.exception(e)
            raise ParseError(self.body, e)

    def wait_for_element(self, css_selector, seconds):
        # ??? -- not sure this is needed or is better than builtin methods
        elem = None
        driver = self.browser
        for count in range(seconds):
            elem = driver.find_element_by_css_selector(css_selector)
            if elem:
                break
            else:
                time.sleep(1)

        return elem


    def __init__(self):
        self.domains = {}

    def save(self):
        """save the browser session for the given domain"""
        logger.debug("Saving cookies for {}".format(self.domain))
        cookie = Cookies(self.domain)
        cookie.save(self.browser.get_cookies())

    def close(self):
        """quit the browser and power down the virtual display"""
        logger.debug("Closing down browser")
        try:
            self.browser.close()
        except Exception as e:
            logger.warn("Browser close failed with {}".format(e.message))
            pass

        logger.debug("Shutting down display")
        self.display.stop()




# -*- coding: utf-8 -*-

import sys

# shamelessly ripped from https://github.com/kennethreitz/requests/blob/master/requests/compat.py
# Syntax sugar.
_ver = sys.version_info
is_py2 = (_ver[0] == 2)
is_py3 = (_ver[0] == 3)

if is_py2:
    from StringIO import StringIO

    basestring = basestring
    range = xrange # range is now always an iterator

    # http://stackoverflow.com/questions/14503751
    # ripped from six https://bitbucket.org/gutworth/six
    exec("""def reraise(tp, value, tb=None):
        try:
            raise tp, value, tb
        finally:
            tb = None
    """)


elif is_py3:
    from io import StringIO

    basestring = (str, bytes)

    # ripped from six https://bitbucket.org/gutworth/six
    def reraise(tp, value, tb=None):
        try:
            if value is None:
                value = tp()
            if value.__traceback__ is not tb:
                raise value.with_traceback(tb)
            raise value
        finally:
            value = None
            tb = None



#!/usr/bin/env python
# http://docs.python.org/distutils/setupscript.html
# http://docs.python.org/2/distutils/examples.html

from setuptools import setup, find_packages
import re
import os


name = "wishlist"
with open(os.path.join(name, "__init__.py"), 'rU') as f:
    version = re.search("^__version__\s*=\s*[\'\"]([^\'\"]+)", f.read(), flags=re.I | re.M).group(1)

setup(
    name=name,
    version=version,
    description='Amazon wishlist scraper',
    author='Jay Marcyes',
    author_email='jay@marcyes.com',
    url='http://github.com/Jaymon/{}'.format(name),
    packages=[name],
    license='GPLv2+',
    install_requires=['captain', 'beautifulsoup4', 'selenium', 'pyvirtualdisplay'],
    classifiers=[ # https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
    ],
    entry_points = {
        'console_scripts': [
            '{} = {}.__main__:console'.format(name, name),
        ],
    },
)



from __future__ import unicode_literals

from captain import echo, exit as console, ArgError
from captain.decorators import arg, args

from wishlist import __version__
from wishlist.core import Wishlist, ParseError
from wishlist.browser import RecoverableCrash


def main_auth():
    """Signin to amazon so you can access private wishlists"""
    with Wishlist.open() as w:
        echo.out("Requesting amazon.com")
        w.homepage(ignore_cookies=True)

        # If you access from another country, amazon might prompt to redirect to
        # country specific store, we don't want that
        if w.element_exists("#redir-opt-out"):
            echo.out("Circumventing redirect")
            remember = w.element("#redir-opt-out")
            stay = w.element("#redir-stay-at-www")
            remember.click()
            stay.click()

        button = w.element("#a-autoid-0-announce")
        echo.out("Clicking sign in button")
        button.click()

        # now put in your creds
        email = w.element("#ap_email")
        password = w.element("#ap_password")
        submit = w.element("#signInSubmit")
        if email and password and submit:
            echo.out("Found sign in form")
            email_in = echo.prompt("Amazon email address")
            password_in = echo.prompt("Amazon password")

        email.send_keys(email_in)
        password.send_keys(password_in)
        echo.out("Signing in")
        submit.click()

        # for 2-factor, wait for this element
        code = w.wait_for_element("#auth-mfa-otpcode", 5)
        if code:
            echo.out("2-Factor authentication is on, you should be receiving a text")
            submit = w.element("#auth-signin-button")
            remember = w.element("#auth-mfa-remember-device")
            remember.click()
            authcode = echo.prompt("2-Factor authcode")
            code.send_keys(authcode)
            submit.click()

        #https://www.amazon.com/ref=gw_sgn_ib/853-0204854-22247543
        if "/ref=gw_sgn_ib/" in w.current_url:
            echo.out("Success, you are now signed in")
            w.save()


@arg('name', help="the name of the wishlist, amazon.com/gp/registry/wishlist/NAME")
@arg('--page', dest="current_page", type=int, default=0, help="The Wishlist page you want to start on")
def main_dump(name, current_page):
    """This is really here just to test that I can parse a wishlist completely and
    to demonstrate (by looking at the code) how to iterate through a list"""
    crash_count = 0
    max_crash_count = 5
    while crash_count < max_crash_count:
        try:
            with Wishlist.open() as w:
                current_url = ""
                for i, item in enumerate(w.get(name, current_page), 1):
                    if current_url:
                        if w.current_url != current_url:
                            current_url = w.current_url
                            echo.h3(current_url)
                    else:
                        current_url = w.current_url
                        echo.h3(current_url)

                    try:
                        item_json = item.jsonable()
                        echo.out("{}. {} is ${:.2f}", i, item_json["title"], item_json["price"])
                        echo.indent(item_json["url"])

                    except ParseError as e:
                        echo.err("{}. Failed!", i)
                        echo.err(e.body)
                        echo.exception(e)

                    except KeyboardInterrupt:
                        raise

                    except Exception as e:
                        echo.err("{}. Failed!", i)
                        echo.exception(e)

                    finally:
                        current_page = w.current_page

        except KeyboardInterrupt:
            break

        except RecoverableCrash:
            crash_count += 1
            if crash_count > max_crash_count:
                raise

        else:
            echo.out("Done with wishlist, {} total pages", current_page)
            break

if __name__ == "__main__":
    console()

